@@ -32,8 +32,10 @@ import io
 IS_PY2 = (sys.version_info[0] == 2)
 
 if IS_PY2:
+    import cStringIO as StringIO
     import httplib
 elif sys.version_info[0] == 3:
+    import io as StringIO
     import http.client as httplib
 
 from boto.compat import BytesIO, urlsplit, six
@@ -623,11 +625,11 @@ def compression_wrapper(file_obj, filename, mode):
             from bz2file import BZ2File
         else:
             from bz2 import BZ2File
-        return make_closing(BZ2File)(filename, mode)
+        return make_closing(BZ2File)(file_obj, mode)
 
     elif ext == '.gz':
         from gzip import GzipFile
-        return make_closing(GzipFile)(filename, mode)
+        return make_closing(GzipFile)(fileobj=file_obj, mode=mode)
 
     else:
         return file_obj
@@ -658,11 +660,6 @@ class HttpReadStream(object):
 
         If none of those are set, will connect unauthenticated.
         """
-        if IS_PY2:
-            from urllib2 import urlopen
-        else:
-            from urllib.request import urlopen
-
         if kerberos:
             import requests_kerberos
             auth = requests_kerberos.HTTPKerberosAuth()
@@ -676,6 +673,7 @@ class HttpReadStream(object):
         if not self.response.ok:
             self.response.raise_for_status()
 
+        self.mode = mode
         self._read_buffer = None
         self._read_iter = None
         self._readline_iter = None
@@ -683,6 +681,10 @@ class HttpReadStream(object):
     def __iter__(self):
         return self.response.iter_lines()
 
+    def binary_content(self):
+        """Return the content of the request as bytes."""
+        return self.response.content
+
     def readline(self):
         """
         Mimics the readline call to a filehandle object.
@@ -693,7 +695,8 @@ class HttpReadStream(object):
         try:
             return next(self._readline_iter)
         except StopIteration:
-            raise EOFError()
+            # When readline runs out of data, it just returns an empty string
+            return ''
 
     def readlines(self):
         """
@@ -701,6 +704,9 @@ class HttpReadStream(object):
         """
         return list(self.response.iter_lines())
 
+    def seek(self):
+        raise NotImplementedError('seek() is not implemented')
+
     def read(self, size=None):
         """
         Mimics the read call to a filehandle object.
@@ -720,7 +726,8 @@ class HttpReadStream(object):
                     retval = self._read_buffer
                     self._read_buffer = ''
                     if len(retval) == 0:
-                        raise EOFError()
+                        # When read runs out of data, it just returns empty
+                        return ''
                     else:
                         return retval
             
@@ -748,7 +755,13 @@ def HttpOpenRead(parsed_uri, mode='r', **kwargs):
     response = HttpReadStream(url, **kwargs)
 
     fname = url.split('/')[-1]
-    return compression_wrapper(response, fname, mode)
+
+    if fname.endswith('.gz'):
+        #  Gzip needs a seek-able filehandle, so we need to buffer it.
+        buffer =  make_closing(io.BytesIO)(response.binary_content())
+        return compression_wrapper(buffer, fname, mode)
+    else:
+        return compression_wrapper(response, fname, mode)
 
 
 class S3OpenWrite(object):
@@ -936,6 +949,10 @@ class WebHdfsOpenWrite(object):
         self.close()
 
 
+def s3_iter_bucket_process_key_with_kwargs(kwargs):
+    return s3_iter_bucket_process_key(**kwargs)
+
+
 def s3_iter_bucket_process_key(key, retries=3):
     """
     Conceptually part of `s3_iter_bucket`, but must remain top-level method because
@@ -955,7 +972,7 @@ def s3_iter_bucket_process_key(key, retries=3):
             pass
 
 
-def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=None, workers=16):
+def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=None, workers=16, retries=3):
     """
     Iterate and download all S3 files under `bucket/prefix`, yielding out
     `(key, key content)` 2-tuples (generator).
@@ -984,15 +1001,15 @@ def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=Non
 
     """
     total_size, key_no = 0, -1
-    keys = (key for key in bucket.list(prefix=prefix) if accept_key(key.name))
+    keys = ({'key': key, 'retries': retries} for key in bucket.list(prefix=prefix) if accept_key(key.name))
 
     if MULTIPROCESSING:
         logger.info("iterating over keys from %s with %i workers" % (bucket, workers))
         pool = multiprocessing.pool.Pool(processes=workers)
-        iterator = pool.imap_unordered(s3_iter_bucket_process_key, keys)
+        iterator = pool.imap_unordered(s3_iter_bucket_process_key_with_kwargs, keys)
     else:
         logger.info("iterating over keys from %s without multiprocessing" % bucket)
-        iterator = imap(s3_iter_bucket_process_key, keys)
+        iterator = imap(s3_iter_bucket_process_key_with_kwargs, keys)
 
     for key_no, (key, content) in enumerate(iterator):
         if key_no % 1000 == 0:
